"""Example sandbox environment server generated by `orwd init`."""
from openreward.environments import Server

from sandbox_env import SandboxEnv

if __name__ == "__main__":
    server = Server([SandboxEnv])
    server.run()
